//==============================================================================
//	
//	Copyright (c) 2002-
//	Authors:
//	* Dave Parker <david.parker@comlab.ox.ac.uk> (University of Oxford)
//	
//------------------------------------------------------------------------------
//	
//	This file is part of PRISM.
//	
//	PRISM is free software; you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation; either version 2 of the License, or
//	(at your option) any later version.
//	
//	PRISM is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//	
//	You should have received a copy of the GNU General Public License
//	along with PRISM; if not, write to the Free Software Foundation,
//	Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//	
//==============================================================================

package explicit;

import java.util.*;
import java.util.function.IntConsumer;

import explicit.nxt.IntDoubleConsumer;
import explicit.nxt.SparseBitSet;
import it.unimi.dsi.fastutil.ints.*;
import prism.PrismUtils;

/**
 * Explicit representation of a probability distribution.
 * Basically, a mapping from (integer-valued) indices to (non-zero, double-valued) probabilities. 
 */
public class Distribution implements Iterable<Int2DoubleMap.Entry> {
	private static final int[] INTS = new int[]{};
	private static final double[] DOUBLES = new double[]{};

	int[] keys;
	double[] values;

	/**
	 * Create an empty distribution.
	 */
	public Distribution()
	{
		keys = INTS;
		values = DOUBLES;
	}

	public Distribution(int key, double value)
	{
		keys = new int[]{key};
		values = new double[]{value};
	}

	/**
	 * Copy constructor.
	 */
	public Distribution(Distribution distr)
	{
		keys = distr.keys.clone();
		values = distr.values.clone();
	}

	/**
	 * Construct a distribution from an existing one and an index permutation,
	 * i.e. in which index i becomes index permut[i].
	 * Note: have to build the new distributions from scratch anyway to do this,
	 * so may as well provide this functionality as a constructor.
	 */
	public Distribution(Distribution distr, int permut[])
	{
		throw new UnsupportedOperationException();
	}

	/**
	 * Clear all entries of the distribution.
	 */
	public void clear()
	{
		keys = INTS;
		values = DOUBLES;
	}

	/**
	 * Add 'prob' to the probability for index 'j'.
	 * Return boolean indicating whether or not there was already
	 * non-zero probability for this index (i.e. false denotes new transition).
	 */
	public boolean add(int j, double prob)
	{
		if (j < 0 || prob <= 0) {
			throw new IllegalArgumentException();
		}

		int i = Arrays.binarySearch(keys, j);

		if (i >= 0) {
			values[i] += prob;
			return true;
		}

		// Compute insertion point
		i = -(i + 1);

		int[] localKeys = new int[keys.length + 1];
		double[] localValues = new double[keys.length + 1];

		System.arraycopy(keys, 0, localKeys, 0, i);
		System.arraycopy(values, 0, localValues, 0, i);
		localKeys[i] = j;
		localValues[i] = prob;
		System.arraycopy(keys, i, localKeys, i + 1, keys.length - i);
		System.arraycopy(values, i, localValues, i + 1, keys.length - i);

		keys = localKeys;
		values = localValues;

		return false;
	}

	/**
	 * Set the probability for index 'j' to 'prob'.
	 */
	public void set(int j, double prob)
	{
		throw new UnsupportedOperationException();
	}

	/**
	 * Get the probability for index j. 
	 */
	public double get(int j)
	{
		return values[Arrays.binarySearch(keys, j)];
	}

	/**
	 * Returns true if index j is in the support of the distribution. 
	 */
	public boolean contains(int j)
	{
		return Arrays.binarySearch(keys, j) >= 0;
	}

	/**
	 * Returns true if all indices in the support of the distribution are in the set. 
	 */
	public boolean isSubsetOf(BitSet set)
	{
		for (int key : keys) {
			if (!set.get(key)) {
				return false;
			}
		}

		return true;
	}

	public boolean isSubsetOf(SparseBitSet set)
	{
		for (int key : keys) {
			if (!set.get(key)) {
				return false;
			}
		}

		return true;
	}

	/**
	 * Returns true if at least one index in the support of the distribution is in the set. 
	 */
	public boolean containsOneOf(BitSet set)
	{
		for (int key : keys) {
			if (set.get(key)) {
				return true;
			}
		}

		return false;
	}

	/**
	 * Get the support of the distribution.
	 */
	public int[] getSupport()
	{
		return keys;
	}

	/**
	 * Get an iterator over the entries of the map defining the distribution.
	 */
	public Iterator<Int2DoubleMap.Entry> iterator()
	{
		return new Iterator<Int2DoubleMap.Entry>() {
			DistributionEntry entry = new DistributionEntry();

			@Override
			public boolean hasNext() {
				return (entry.index + 1) < keys.length;
			}

			@Override
			public Int2DoubleMap.Entry next() {
				entry.index++;
				return entry;
			}
		};
	}

	/**
	 * Returns true if the distribution is empty.
	 */
	public boolean isEmpty()
	{
		return keys.length == 0;
	}

	/**
	 * Get the size of the support of the distribution.
	 */
	public int size()
	{
		return keys.length;
	}

	/**
	 * Get the sum of the probabilities in the distribution.
	 */
	public double sum()
	{
		double d = 0.0;

		for (double prob : values) {
			d  += prob;
		}

		return d;
	}

	public double sumWeighted(double[] vect) {
		double d = 0.0;
		int i = 0;

		for (int key : keys) {
			d += vect[key] * values[i];
		}

		return d;
	}

	/**
	 * Get the sum of all the probabilities in the distribution except for index j.
	 */
	public double sumAllBut(int j)
	{
		throw new UnsupportedOperationException();
	}

	/**
	 * Create a new distribution, based on a mapping from the indices
	 * used in this distribution to a different set of indices.
	 */
	public Distribution map(int map[])
	{
		throw new UnsupportedOperationException();
	}

	@Override
	public boolean equals(Object o) {
		if (this == o) return true;
		if (o == null || getClass() != o.getClass()) return false;
		Distribution that = (Distribution) o;
		return Arrays.equals(keys, that.keys) &&
				PrismUtils.doublesAreClose(values, that.values, 1e-12, false);
	}

	@Override
	public int hashCode()
	{
		return Arrays.hashCode(keys);
	}

	@Override
	public String toString()
	{
		return "{" + keys.toString() + " -> " + values.toString() + "}";
	}

	public void forEach(IntDoubleConsumer consumer) {
		int i = 0;
		for (int key : keys) {
			consumer.accept(key, values[i]);
			i++;
		}
	}

	public void forEach(IntConsumer consumer) {
		for (int key : keys) {
			consumer.accept(key);
		}
	}

	class DistributionEntry implements Int2DoubleMap.Entry {
		int index = -1;

		@Override
        public Integer getKey() {
            return getIntKey();
        }

		@Override
        public int getIntKey() {
            return keys[index];
        }

		@Override
        public Double getValue() {
            return getDoubleValue();
        }

		@Override
        public double setValue(double value) {
            throw new UnsupportedOperationException();
        }

		@Override
        public double getDoubleValue() {
            return values[index];
        }

		@Override
        public Double setValue(Double value) {
			throw new UnsupportedOperationException();
        }
	}
}
