/*
 * Copyright (C) 2016  (Salomon Sickert)
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package explicit.nxt;

import parser.ast.*;
import parser.type.TypeBool;
import prism.PrismLangException;
import rabinizer.ltl.*;

/*
 * Convert a property expression (an LTL formula) to classes understood by ltl2ldba
 */
public class ExpressionToLTL {

    public static Formula convert(Expression e) throws PrismLangException {
        if (e == null) {
            return null;
        }

        if (e instanceof ExpressionTemporal) {
            return convertTemporal((ExpressionTemporal) e);
        } else if (e instanceof ExpressionBinaryOp) {
            return convertBinaryOp((ExpressionBinaryOp) e);
        } else if (e instanceof ExpressionUnaryOp) {
            return convertUnaryOp((ExpressionUnaryOp) e);
        } else if (e instanceof ExpressionLiteral) {
            return convertLiteral((ExpressionLiteral) e);
        } else if (e instanceof ExpressionLabel) {
            return convertLabel((ExpressionLabel) e);
        }

        throw new IllegalArgumentException("Cannot convert expression: " + e.toString());
    }

    private static Formula convertTemporal(ExpressionTemporal e) throws PrismLangException {
        Formula ltl1 = convert(e.getOperand1());
        Formula ltl2 = convert(e.getOperand2());

        switch (e.getOperator()) {
            case ExpressionTemporal.P_X:
                return new XOperator(ltl2);

            case ExpressionTemporal.P_U:
                return new UOperator(ltl1, ltl2);

            case ExpressionTemporal.P_F:
                return new FOperator(ltl2);

            case ExpressionTemporal.P_G:
                return new GOperator(ltl2);

            case ExpressionTemporal.P_W:
            case ExpressionTemporal.P_R:
                return convert(e.convertToUntilForm());

            default:
                throw new PrismLangException("Cannot convert expression: ", e);
        }
    }

    private static Formula convertBinaryOp(ExpressionBinaryOp e) throws PrismLangException {
        Formula ltl1 = convert(e.getOperand1());
        Formula ltl2 = convert(e.getOperand2());

        switch (e.getOperator()) {
            case ExpressionBinaryOp.IMPLIES:
                return new Disjunction(ltl1.not(), ltl2);

            case ExpressionBinaryOp.IFF:
                return new Conjunction(new Disjunction(ltl1.not(), ltl2), new Disjunction(ltl1, ltl2.not()));

            case ExpressionBinaryOp.OR:
                return new Disjunction(ltl1, ltl2);

            case ExpressionBinaryOp.AND:
                return new Conjunction(ltl1, ltl2);

            default:
                throw new PrismLangException("Cannot convert expression: ", e);
        }
    }

    private static Formula convertUnaryOp(ExpressionUnaryOp e) throws PrismLangException {
        Formula ltl1 = convert(e.getOperand());

        switch (e.getOperator()) {
            case ExpressionUnaryOp.NOT:
                return ltl1.not();

            case ExpressionUnaryOp.PARENTH:
                return ltl1;

            default:
                throw new PrismLangException("Cannot convert expression: ", e);
        }
    }

    private static Formula convertLiteral(ExpressionLiteral e) throws PrismLangException {
        if (!(e.getType() instanceof TypeBool)) {
            throw new PrismLangException("Cannot convert expression: ", e);
        }

        return BooleanConstant.get(e.evaluateBoolean());
    }

    private static Formula convertLabel(ExpressionLabel e) throws PrismLangException {
        return new Literal(Integer.parseInt(e.getName()), false);
    }
}
