/*
 * Copyright (C) 2016
 *
 * Dave Parker <david.parker@comlab.ox.ac.uk> (University of Oxford)
 * Joachim Klein <klein@tcs.inf.tu-dresden.de> (TU Dresden)
 * Salomon Sickert <sickert@in.tum.de> (TU München)
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package explicit.nxt;

import explicit.*;
import explicit.Model;
import explicit.ProbModelChecker;
import explicit.StateModelChecker;
import explicit.StateValues;
import jhoafparser.consumer.HOAConsumerException;
import jhoafparser.consumer.HOAConsumerPrint;
import parser.ast.*;
import parser.type.TypeBool;
import parser.type.TypePathBool;
import prism.*;
import rabinizer.automata.Optimisation;
import ltl2ldba.DetLimitAutomaton;
import ltl2ldba.DetLimitAutomatonFactory;
import rabinizer.ltl.Formula;

import java.io.*;
import java.util.*;
import java.util.List;

/**
 * LTL model checking functionality, Based on MDPModelChecker
 */
public class LDBALTLModelChecker extends PrismComponent
{

	/**
	 * Create a new LTLModelChecker, inherit basic state from parent (unless null).
	 */
	public LDBALTLModelChecker(PrismComponent parent)
	{
		super(parent);
	}

	/**
	 * Extract maximal state formula from an LTL path formula, model check them (with passed in model checker) and
	 * replace them with ExpressionLabel objects L0, L1, etc. Expression passed in is modified directly, but the result
	 * is also returned. As an optimisation, expressions that results in true/false for all states are converted to an
	 * actual true/false, and duplicate results (or their negations) reuse the same label. BitSets giving the states which
	 * satisfy each label are put into the vector labelBS, which should be empty when this function is called.
	 */
	public Expression checkMaximalStateFormulas(StateModelChecker mc, Model model, Expression expr, List<BitSet> labelBS) throws PrismException
	{
		// A state formula
		if (expr.getType() instanceof TypeBool) {
			// Model check state formula for all states
			StateValues sv = mc.checkExpression(model, expr, null);
			BitSet bs = sv.getBitSet();
			// Detect special cases (true, false) for optimisation
			if (bs.isEmpty()) {
				return Expression.False();
			}
			if (bs.cardinality() == model.getNumStates()) {
				return Expression.True();
			}
			// See if we already have an identical result
			// (in which case, reuse it)
			int i = labelBS.indexOf(bs);
			if (i != -1) {
				sv.clear();
				return new ExpressionLabel(Integer.toString(i));
			}
			// Also, see if we already have the negation of this result
			// (in which case, reuse it)
			BitSet bsNeg = new BitSet(model.getNumStates());
			bsNeg.set(0, model.getNumStates());
			bsNeg.andNot(bs);
			i = labelBS.indexOf(bsNeg);
			if (i != -1) {
				sv.clear();
				return Expression.Not(new ExpressionLabel(Integer.toString(i)));
			}
			// Otherwise, add result to list, return new label
			labelBS.add(bs);
			return new ExpressionLabel(Integer.toString(labelBS.size() - 1));
		}
		// A path formula (recurse, modify, return)
		else if (expr.getType() instanceof TypePathBool) {
			if (expr instanceof ExpressionBinaryOp) {
				ExpressionBinaryOp exprBinOp = (ExpressionBinaryOp) expr;
				exprBinOp.setOperand1(checkMaximalStateFormulas(mc, model, exprBinOp.getOperand1(), labelBS));
				exprBinOp.setOperand2(checkMaximalStateFormulas(mc, model, exprBinOp.getOperand2(), labelBS));
			} else if (expr instanceof ExpressionUnaryOp) {
				ExpressionUnaryOp exprUnOp = (ExpressionUnaryOp) expr;
				exprUnOp.setOperand(checkMaximalStateFormulas(mc, model, exprUnOp.getOperand(), labelBS));
			} else if (expr instanceof ExpressionTemporal) {
				ExpressionTemporal exprTemp = (ExpressionTemporal) expr;
				if (exprTemp.getOperand1() != null) {
					exprTemp.setOperand1(checkMaximalStateFormulas(mc, model, exprTemp.getOperand1(), labelBS));
				}
				if (exprTemp.getOperand2() != null) {
					exprTemp.setOperand2(checkMaximalStateFormulas(mc, model, exprTemp.getOperand2(), labelBS));
				}
			}
		}
		return expr;
	}

	/**
	 * Construct a deterministic automaton (DA) for an LTL formula, having first extracted maximal state formulas
	 * and model checked them with the passed in model checker. The maximal state formulas are assigned labels
	 * (L0, L1, etc.) which become the atomic propositions in the resulting DA. BitSets giving the states which
	 * satisfy each label are put into the vector {@code labelBS}, which should be empty when this function is called.
	 *
	 * @param mc a ProbModelChecker, used for checking maximal state formulas
	 * @param model the model
	 * @param expr a path expression, i.e. the LTL formula
	 * @param labelBS empty vector to be filled with BitSets for subformulas
	 * @return the DA
	 */
	private LDBAWrapper constructDAForLTLFormula(ProbModelChecker mc, MDP model, Expression expr, List<BitSet> labelBS) throws PrismException
	{
		Expression ltl;
		DetLimitAutomaton da;
		long time;

		if (Expression.containsTemporalTimeBounds(expr)) {
			if (model.getModelType().continuousTime()) {
				throw new PrismException("Automaton construction for time-bounded operators not supported for " + model.getModelType()+".");
			}

			if (!expr.isSimplePathFormula()) {
				throw new PrismNotSupportedException("Time-bounded operators not supported in LTL: " + expr);
			}
		}

		// Model check maximal state formulas
		ltl = checkMaximalStateFormulas(mc, model, expr.deepCopy(), labelBS);

		// Convert LTL formula to deterministic automaton
		mainLog.println("\nBuilding deterministic automaton (for " + ltl + ")...");
		Formula ltlFormula = ExpressionToLTL.convert(ltl);
		EnumSet<Optimisation> optimisations = EnumSet.allOf(Optimisation.class);
		optimisations.remove(Optimisation.REMOVE_EPSILON_TRANSITIONS);
		optimisations.remove(Optimisation.LAZY_ACCEPTING_COMPONENT_CONSTRUCTION);
		time = System.currentTimeMillis();

		da = DetLimitAutomatonFactory.createDetLimitAutomaton(ltlFormula, null, optimisations);

		mainLog.println("A has at least " + da.size() + " states. The initial state is " + da.getInitialState());
		time = System.currentTimeMillis() - time;
		mainLog.println("Time for translation: " + time / 1000.0 + " seconds.");
		// If required, export DA
		if (settings != null && settings.getExportPropAut()) {
			mainLog.println("Exporting to file \"" + settings.getExportPropAutFilename() + "\"...");
			PrintStream out = PrismUtils.newPrintStream(settings.getExportPropAutFilename());
			try {
				da.toHOA(new HOAConsumerPrint(out));
			} catch (HOAConsumerException e) {
				e.printStackTrace();
			}
			out.close();
		}
		
		return new LDBAWrapper(da);
	}

	/**
	 * Generate a deterministic automaton for the given LTL formula
	 * and construct the product of this automaton with an MDP.
	 *
	 * @param mc a ProbModelChecker, used for checking maximal state formulas
	 * @param model the model
	 * @param expr a path expression
	 * @param statesOfInterest the set of states for which values should be calculated (null = all states)
	 * @return the product with the DA
	 * @throws PrismException
	 */
	public LDBAProduct<MDP> constructProductMDP(ProbModelChecker mc, MDP model, Expression expr, BitSet statesOfInterest) throws PrismException
	{
		// Convert LTL formula to automaton
		List<BitSet> labelBS = new ArrayList<>();
		LDBAWrapper da = constructDAForLTLFormula(mc, model, expr, labelBS);

		// Build product of model and automaton
		long startTime = System.nanoTime();
		mainLog.println("\nConstructing MDP-LDBA product and finding accepting MECs...");
		LDBAProduct<MDP> product = LDBAProductBuilder.constructProductModel(da, model, labelBS, statesOfInterest);
		long estimatedTime = System.nanoTime() - startTime;
		System.out.println("Construction Time: " + (estimatedTime * 0.000000001) + " seconds.");
		mainLog.print("\n" + product.getProductModel().infoStringTable());

		return product;
	}
}
