/*
 * Copyright (C) 2016
 *
 * Dave Parker <david.parker@comlab.ox.ac.uk> (University of Oxford)
 * Joachim Klein <klein@tcs.inf.tu-dresden.de> (TU Dresden)
 * Salomon Sickert <sickert@in.tum.de> (TU München)
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package explicit.nxt;

import explicit.*;
import it.unimi.dsi.fastutil.ints.*;
import parser.VarList;
import parser.ast.Declaration;
import parser.ast.DeclarationInt;
import parser.ast.Expression;
import prism.PrismException;
import rabinizer.collections.valuationset.ValuationSet;

import java.awt.*;
import java.util.*;
import java.util.List;

/* Based on LDBAProductBuilder from explicit. */
public final class LDBAProductBuilder {

    private final LDBAWrapper da;
    private final MDP model;
    private final int daSize;
    private final List<BitSet> labelBS;
    private final BitSet statesOfInterest;

    // Encoding:
    // each state s' = <s, q> = s + (modelSize * q)
    // s(s') = s' % modelSize
    // q(s') = s' / modelSize
    private int[] map;
    private int trapState;

    private LDBAProductBuilder(MDP m, LDBAWrapper da, List<BitSet> labelBS, BitSet statesOfInterest) {
        model = m;
        this.da = da;
        daSize = da.size();

        this.labelBS = labelBS;
        this.statesOfInterest = statesOfInterest;
        this.map = new int[m.getNumStates() * da.size()];
        Arrays.fill(map, -1);
        trapState = -1;
    }

    /**
     * Lookup the product state
     * @param s The model state
     * @param q The LDBA state
     * @return
     */
    private int getIndex(int s, int q) {
        return (s * daSize) + q;
    }

    private void buildTrapState(MDPSimple prodModel) {
        trapState = prodModel.addState();
        prodModel.addChoice(trapState, new Distribution(trapState, 1.0d));
    }

    /**
     * Construct the product of a DA and a model.
     * @param da The DA
     * @param model The model
     * @param labelBS BitSets giving the set of states for each AP in the DA
     * @param statesOfInterest the set of states for which values should be calculated (null = all states)
     * @return The product model
     */
    public static LDBAProduct<MDP> constructProductModel(LDBAWrapper da, MDP model, List<BitSet> labelBS, BitSet statesOfInterest) throws PrismException
    {
        LDBAProductBuilder builder = new LDBAProductBuilder(model, da, labelBS, statesOfInterest);
        return builder.build();
    }

    private LDBAProduct<MDP> build() throws PrismException {
        MDPSimple prodModel = new MDPSimple();

        if (model.getVarList() != null) {
            VarList varList = model.getVarList();
            // Create a (new, unique) name for the variable that will represent DA states
            String daVar = "_da";
            while (varList.getIndex(daVar) != -1) {
                daVar = "_" + daVar;
            }

            VarList newVarList = (VarList) varList.clone();
            Declaration decl = new Declaration(daVar, new DeclarationInt(Expression.Int(0), Expression.Int(Math.max(da.size() - 1, 1))));
            newVarList.addVar(0, decl, 1, model.getConstantValues());
            prodModel.setVarList(newVarList);
        }

        Deque<Point> queue = new ArrayDeque<>();

        // We need results for all states of the original model in statesOfInterest
        // We thus explore states of the product starting from these states.
        // These are designated as initial states of the product model
        // (a) to ensure reachability is done for these states; and
        // (b) to later identify the corresponding product state for the original states
        //     of interest

        Int2IntMap initialStateInverseMapping = new Int2IntArrayMap(statesOfInterest.cardinality());

        for (int s0 = statesOfInterest.nextSetBit(0); s0 >= 0; s0 = statesOfInterest.nextSetBit(s0+1)) {
            // Find corresponding initial state in DA
            int q0 = da.getEdgeDestByLabel(LDBAWrapper.getInitialState(), getLabel(s0));

            if (q0 >= 0) {
                int pi = getIndex(s0, q0);
                int p0 = map[pi];

                if (p0 == -1) {
                    p0 = prodModel.addState();
                    map[pi] = p0;
                    queue.add(new Point(s0, q0));
                    prodModel.addInitialState(p0);
                    initialStateInverseMapping.put(p0, s0);
                }
                // TODO: detect "true" trap state.
            } else if (trapState == -1) {
                buildTrapState(prodModel);
                prodModel.addInitialState(trapState);
                initialStateInverseMapping.put(trapState, s0);
            }
        }

        boolean initialComponentPresent = (da.getSmallestAcceptingState() > 0);
        SparseBitSet goalStates = new SparseBitSet();
        constructReachableStates(prodModel, queue, initialComponentPresent ? null : goalStates);

        if (initialComponentPresent) {
            // Build a mapping from state indices to states (s,q), encoded as (s + (modelSize * q))
            int[] invMap = new int[prodModel.getNumStates()];
            for (int i1 = da.getSmallestAcceptingState() - 1; i1 >= 0; i1--) {
                int j = map[i1];

                if (j != -1) {
                    invMap[j] = i1;
                }
            }

            // MECs have to be disjoint!
            for (LightMEC mec : LightECComputer.computeMECs(prodModel, null)) {
                for (int i = mec.states.nextSetBit(0); i >= 0; i = mec.states.nextSetBit(i + 1)) {
                    // Get Model state
                    int s0 = invMap[i] / daSize;
                    // Get LDBA state.
                    int q0 = invMap[i] % daSize;

                    int[] js = da.getJumpTargets(q0);

                    for (int j : js) {
                        int pi = getIndex(s0, j);
                        int p = map[pi];

                        // Add state to model
                        if (p == -1) {
                            p = prodModel.addState();
                            map[pi] = p;
                            queue.add(new Point(s0, j));
                            constructReachableStates(prodModel, queue, goalStates);
                        }

                        // Add transition
                        prodModel.addChoice(i, new Distribution(p, 1.0d));
                    }
                }
            }
        }

        // Clear memory
        map = null;

        // Copy model to the sparse model to improve performance
        return new LDBAProduct<>(new MDPSparse(prodModel), model, initialStateInverseMapping, goalStates);
    }

    // Get BitSet representing APs (labels) satisfied by successor state s
    private BitSet getLabel(int s) {
        BitSet label = new BitSet();

        ListIterator<BitSet> iterator = labelBS.listIterator();

        while (iterator.hasNext()) {
            if (iterator.next().get(s)) {
                label.set(iterator.previousIndex());
            }
        }

        return label;
    }

    private void constructReachableStates(MDPSimple prodModel, Deque<Point> queue, SparseBitSet goalStates) {
        // local acceptance
        List<Int2ObjectMap<BitSet>> buchi = new ArrayList<>();
        SparseBitSet preMEC = null;

        if (goalStates != null) {
            preMEC = new SparseBitSet();
            int size = da.getAcceptanceSize();

            for (int i = 0; i < size; i++) {
                buchi.add(new Int2ObjectLinkedOpenHashMap<>());
            }
        }

        while (!queue.isEmpty()) {
            Point current = queue.remove();

            int s1 = current.x;
            int q1 = current.y;
            int p1 = map[getIndex(s1, q1)];

            Map<BitSet, ValuationSet> acceptance = da.getAcceptance(q1);

            if (preMEC != null) {
                preMEC.set(p1);
            }

            // Go through transitions from state s_1 in original model
            final int numChoices = model.getNumChoices(s1);

            for (int j = 0; j < numChoices; j++) {
                Distribution prodDistr = new Distribution();
                BitSet accIndexes = new BitSet();

                for (Int2DoubleMap.Entry e : model.getTransitions(s1, j)) {
                    int s2 = e.getIntKey();
                    BitSet label = getLabel(s2);
                    int q2 = da.getEdgeDestByLabel(q1, label);
                    int p2;

                    if (q2 >= 0) {
                        int pi = getIndex(s2, q2);
                        p2 = map[pi];

                        // Add state to model
                        if (p2 == -1) {
                            p2 = prodModel.addState();
                            map[pi] = p2;
                            queue.add(new Point(s2, q2));
                        }

                        // If there is an acceptance mapping, select all accepting edges
                        if (acceptance != null) {
                            acceptance.forEach((k, v) -> {
                                if (v.contains(label)) {
                                    accIndexes.or(k);
                                }
                            });
                        }
                    } else if (trapState == -1) {
                        buildTrapState(prodModel);
                        p2 = trapState;
                    } else {
                        p2 = trapState;
                    }

                    // Add transition
                    prodDistr.add(p2, e.getDoubleValue());
                }

                for (int i = accIndexes.nextSetBit(0); i >= 0; i = accIndexes.nextSetBit(i+1)) {
                    Int2ObjectMap<BitSet> map = buchi.get(i);
                    BitSet actions = map.get(p1);

                    if (actions == null) {
                        actions = new BitSet();
                        map.put(p1, actions);
                    }

                    actions.set(j);
                }

                prodModel.addChoice(p1, prodDistr);
            }
        }

        if (goalStates != null && !preMEC.isEmpty()) {
            SparseBitSet localGoalStates = LightAcceptingEC.findAcceptingECStates(prodModel, buchi, Collections.singleton(LightMEC.createMEC(prodModel, preMEC)));
            goalStates.or(localGoalStates);
        }
    }
}
