/*
 * Copyright (C) 2016 (Salomon Sickert)
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package explicit.nxt;

import it.unimi.dsi.fastutil.ints.Int2ObjectMap;
import it.unimi.dsi.fastutil.ints.Int2ObjectOpenHashMap;
import it.unimi.dsi.fastutil.objects.Object2IntMap;
import it.unimi.dsi.fastutil.objects.Object2IntOpenHashMap;
import ltl2ldba.AcceptingComponent;
import ltl2ldba.DetLimitAutomaton;
import rabinizer.automata.IState;
import rabinizer.automata.Master;
import rabinizer.collections.valuationset.ValuationSet;

import java.util.*;

public class LDBAWrapper {

    private static final int[] EMPTY = new int[0];

    private final DetLimitAutomaton a;
    private final Object2IntMap<IState<?>> numbering;
    private final List<IState<?>> map;
    private final Int2ObjectMap<int[]> jumpCache;
    private final int smallestAcceptingState;

    public LDBAWrapper(DetLimitAutomaton a) {
        this.a = a;

        // The non-accepting trap state is stored as -1.
        numbering = new Object2IntOpenHashMap<>(a.size());
        numbering.defaultReturnValue(-2);
        map = new ArrayList<>(a.size());

        // Register initial state as state 0;
        get(a.getInitialState());

        if (a.initialComponent != null) {
            a.initialComponent.getStates().forEach(this::get);
            smallestAcceptingState = map.size();
        } else {
            smallestAcceptingState = 0;
        }

        a.acceptingComponent.getStates().forEach(this::get);
        jumpCache = new Int2ObjectOpenHashMap<>();
    }

    public Map<BitSet, ValuationSet> getAcceptance(int i) {
        IState<?> state = get(i);

        if (state instanceof AcceptingComponent.State) {
            AcceptingComponent.State s = (AcceptingComponent.State) state;
            return s.getAcceptance();
        }

        return null;
    }

    public int getAcceptanceSize() {
        return a.acceptingComponent.getAcceptanceSize();
    }

    /**
     * Get the destination of the edge from state i with label lab
     * @param i state
     * @param label the label of the edge
     * @return Returns -1 if no such edge is found.
     */
    public int getEdgeDestByLabel(int i, BitSet label){
        IState<?> s = get(i);

        if (s == null) {
            return -1;
        }

        if (s instanceof Master.State) {
            s = a.initialComponent.getSuccessor((Master.State) s, label);
        } else {
            s = a.acceptingComponent.getSuccessor((AcceptingComponent.State) s, label);
        }

        return get(s);
    }

    public static int getInitialState() {
        return 0;
    }

    public int[] getJumpTargets(int i) {
        int[] result = jumpCache.get(i);

        if (result != null) {
            return result;
        }

        IState<?> state = get(i);

        if (state instanceof Master.State) {
            Set<AcceptingComponent.State> targets = a.initialComponent.epsilonJumps.get((Master.State) state);

            result = new int[targets.size()];
            int c = 0;
            for (AcceptingComponent.State target : targets) {
                result[c] = get(target);
                c++;
            }

            jumpCache.put(i, result);

            return result;
        }

        return EMPTY;
    }

    public int getSmallestAcceptingState() {
        return smallestAcceptingState;
    }

    public int size() {
        return a.size();
    }

    @Override
    public String toString() {
        return a.toString();
    }

    private IState<?> get(int i) {
        return map.get(i);
    }

    private int get(IState<?> state) {
        if (state == null) {
            return -1;
        }

        int i = numbering.getInt(state);

        if (i == -2) {
            i = numbering.size();
            map.add(state);
            numbering.put(state, i);
        }

        return i;
    }
}
