/*
 * Copyright (C) 2016  (Salomon Sickert)
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package explicit.nxt;

import explicit.MDP;

import java.util.*;

/**
 * Explicit maximal end component computer for a nondeterministic model such as an MDP.
 * Implements a variation of the algorithm from p.48 of:
 * Luca de Alfaro. Formal Verification of Probabilistic Systems. Ph.D. thesis, Stanford University (1997)
 */
public class LightECComputer {

	/**
	 * Compute MECs for the given MDP.
	 * @param model
	 * @param preMECs if null all reachable states are considered.
     * @return A list of all MECs obtained from refining the preMECs.
     */
	public static List<LightMEC> computeMECs(MDP model, Collection<LightMEC> preMECs) {
		if (preMECs == null) {
			SparseBitSet bs = new SparseBitSet();
			bs.set(0, model.getNumStates());
			preMECs = new ArrayList<>();
			preMECs.add(LightMEC.createMEC(model, bs));
		}

		List<LightMEC> mecs = new ArrayList<>();
		Deque<LightMEC> workList = new ArrayDeque<>(preMECs);

		while (!workList.isEmpty()) {
			LightMEC mec = workList.remove();
			List<LightMEC> refinedMECs = LightSCCComputerTarjan.refineMEC(model, mec);

			if (!refinedMECs.isEmpty()) {
				LightMEC refinedMEC = refinedMECs.get(0);

				if (mec.equals(refinedMEC)) {
					mecs.add(refinedMEC);
				} else {
					workList.addAll(refinedMECs);
				}
			}
		}

		return mecs;
	}
}
