/*
 * Copyright (C) 2016  (Salomon Sickert)
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package explicit.nxt;

import explicit.Distribution;
import explicit.MDPSimple;
import explicit.NondetModel;
import it.unimi.dsi.fastutil.ints.*;

import java.util.*;

/**
 * LightMEC describes a MEC for a give model. It does not store the model itself. Only the states belonging to the MEC and
 * the allowed actions.
 */
public class LightMEC {
    public final SparseBitSet states;
    public final Int2ObjectMap<BitSet> actions;
    
    private LightMEC(SparseBitSet s, Int2ObjectMap<BitSet> a) {
        states = s;
        actions = a;
    }

    public static LightMEC createMEC(MDPSimple model, SparseBitSet states) {
        Int2ObjectMap<BitSet> actions = new Int2ObjectOpenHashMap<>(states.cardinality());
        boolean changed = true;

        while (changed) {
            changed = false;

            for (int i = states.nextSetBit(0); i >= 0; i = states.nextSetBit(i + 1)) {
                BitSet act = actions.get(i);

                Distribution[] distributions = model.trans.get(i);

                if (distributions != null) {
                    if (act == null) {
                        act = new BitSet();
                        act.set(0, distributions.length);
                        actions.put(i, act);
                    }

                    for (int j = act.nextSetBit(0); j >= 0; j = act.nextSetBit(j + 1)) {
                        if (!distributions[j].isSubsetOf(states)) {
                            act.clear(j);
                        }
                    }
                }

                if (act == null || act.isEmpty()) {
                    states.clear(i);
                    actions.remove(i);
                    changed = true;
                }
            }
        }

        return new LightMEC(states, actions);
    }

    public static LightMEC createMEC(NondetModel model, SparseBitSet states) {
        if (model instanceof MDPSimple) {
            return createMEC((MDPSimple) model, states);
        }

        Int2ObjectMap<BitSet> actions = new Int2ObjectOpenHashMap<>(states.cardinality());
        boolean changed = true;

        while (changed) {
            changed = false;

            for (int i = states.nextSetBit(0); i >= 0; i = states.nextSetBit(i + 1)) {
                BitSet act = actions.get(i);

                final int choices = model.getNumChoices(i);

                if (act == null) {
                    act = new BitSet();
                    act.set(0, choices);
                    actions.put(i, act);
                }

                for (int j = act.nextSetBit(0); j >= 0; j = act.nextSetBit(j + 1)) {
                    if (!model.allSuccessorsInSet(i, j, states)) {
                        act.clear(j);
                    }
                }

                if (act.isEmpty()) {
                    states.clear(i);
                    actions.remove(i);
                    changed = true;
                }
            }
        }

        return new LightMEC(states, actions);
    }

    @Override
    public boolean equals(Object o) {
        if (this == o) return true;
        if (o == null || getClass() != o.getClass()) return false;
        LightMEC lightMEC = (LightMEC) o;
        return Objects.equals(states, lightMEC.states);
    }

    @Override
    public int hashCode() {
        return Objects.hash(states);
    }
}
