/*
 * Copyright (C) 2016  (Salomon Sickert)
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package explicit.nxt;

import explicit.MDP;
import it.unimi.dsi.fastutil.ints.*;

import java.util.*;

/**
 * Tarjan's SCC algorithm operating on an MDP.
 */
public class LightSCCComputerTarjan
{
	private final MDP model;
	private final Int2ObjectMap<BitSet> allowedActions;
	private final List<LightMEC> refinedMECs = new ArrayList<>();

	private int index = 0;

	private final IntStack stack;
	private final SparseBitSet onStack;
    private final SparseBitSet todo;
	private final Int2IntMap lowlinks;
	private final Int2IntMap indices;

	private LightSCCComputerTarjan(MDP model, LightMEC mec) {
		this.model = model;
		this.allowedActions = mec.actions;

        todo = mec.states.clone();
		stack = new IntArrayList();
		onStack = new SparseBitSet();
		lowlinks = new Int2IntOpenHashMap();
		lowlinks.defaultReturnValue(-1);
		indices = new Int2IntOpenHashMap();
		indices.defaultReturnValue(-1);
	}

	public static List<LightMEC> refineMEC(MDP model, LightMEC mec) {
		LightSCCComputerTarjan sccc = new LightSCCComputerTarjan(model, mec);
		sccc.refineMEC();
		return sccc.refinedMECs;
	}

	private void refineMEC() {
		for (int i = todo.nextSetBit(0); i >= 0; i = todo.nextSetBit(i+1)) {
			tarjan(i);
		}
		refinedMECs.removeIf(mecs -> mecs.states.isEmpty());
	}

	// TODO: Replace recurive calls by while-loop and an int call stack.
	private void tarjan(int id) {
		int idIndex = index;
		indices.put(id, index);
		lowlinks.put(id, index);
		index++;
		stack.push(id);
		onStack.set(id);
        todo.clear(id);

		BitSet actions = allowedActions.get(id);

		for (int action = actions.nextSetBit(0); action >= 0; action = actions.nextSetBit(action+1)) {
			IntIterator successorIterator = model.getSuccessorsIterator(id, action);

			while (successorIterator.hasNext()) {
				int e = successorIterator.nextInt();

				if (todo.get(e)) {
					tarjan(e);
					int eLowlink = lowlinks.get(e);
					if (eLowlink < lowlinks.get(id)) {
						lowlinks.put(id, eLowlink);
					}
				} else if (onStack.get(e)) {
                    int eIndex = indices.get(e);
                    if (eIndex < lowlinks.get(id)) {
                        lowlinks.put(id, eIndex);
                    }
                }
			}
		}

		if (lowlinks.get(id) == idIndex) {
			int n = -1;
			SparseBitSet scc = new SparseBitSet();

			while (!stack.isEmpty() && n != idIndex) {
				n = stack.popInt();
				lowlinks.remove(n);
                indices.remove(n);
				onStack.clear(n);
				scc.set(n);
			}

			refinedMECs.add(LightMEC.createMEC(model, scc));
		}
	}
}
